/***************************************************
 * fav2bmk.c
 * =========
 * Written by Rob Stockley (robstockley@bigfoot.com)
 * July 1999
 *
 * I recently changed from using Windows most of the time
 * and Linux occasionally to using Linux 95% of the time 
 * and Windows only for cool games like Apache Longbow II.
 *
 * Under Linux I'm using Netscape Communicator for all my 
 * internet work but while under Windows I used IE. When
 * changing I ran into the following problem.  IE stores 
 * all my bookmarks as files in a nest of folders, but
 * Netscape stores them as a single HTML file.  My problem.
 * How do I keep all my bookmarks without having to type
 * them all out in the new format?  My answer, write a program
 * to automate the process.
 *
 * This is my first programming project involving C for
 * quite some time.  I know I could achieve most, if not 
 * all of this with a shell script. However, I'm writing
 * it in C as an exercise before moving on to some more 
 * interesting projects.
 *
 * Copying
 * =======
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Usage
 * =====
 * Prior to starting you must ensure that your windows path
 * is mounted and readable by you. Then invoke fav2bmk as follows.
 *
 * fav2bmk ie_favorites_dir > destination_file
 *  
 *   eg.  fav2bmk /dos/windows/Favorites > ~/new_bookmarks.html
 *
 * In Netscape click Bookmarks->Edit Bookmarks, then File->Import.
 * Select the destination file you gave and voila, you now have a 
 * new bookmarks directory containing all your IE favorites. You
 * can move them around as required.
 *
 * Concept
 * =======
 * Basically the program recursively scans all the files 
 * and subdirectories beneath the path given on the command
 * line.  When it finds a file fitting the mask *.url it 
 * creates a bookmark for it.  Directories are translated 
 * into leaves on a results tree.  Empty directories 
 * are kept. Once the path has been completely scanned
 * the results tree is parsed and the HTML source is output
 * to the stdout ready to be piped to a file.
 *
 * I know there's an awful lot of error checking done throughout
 * this code but in my opinion it's all worth having even if some 
 * of it is redundant.  Programs that don't depend on real-time 
 * loops can only benefit from stronger error checking and the 
 * extra few micro-seconds it takes to run are insignificant
 * on most of todays machines.
 *
 * Bugs
 * ====
 * None that I know of, however, there is conceptually no limit 
 * to how deep the path will be searched. This means that if you 
 * do something like 'fav2bmk /' you might run out of memory, disk
 * space, or both.  That's not really a bug but just something to 
 * be aware of. I don't know if Netscape sets a limit on the maximum 
 * depth of the bookmarks path.  Also the windows path must be mounted
 * readable by you. If not then fav2bmk will report that the path is
 * bad or unreachable.
 *
 * Copyrights etc.
 * ===============
 * It is my intention to announce this in comp.os.linux.announce
 * and if I get any interest from others to release it into 
 * the public domain under GNU General Public Licence.  If you
 * find this program useful or if you've got suggestions or bugs
 * to report then please send me an email with 'fav2bmk' somewhere 
 * in the subject line.  Enjoy!
 * 
 * Thanks To
 * =========
 * Adam Jenkins - for advice regarding GPL and bug reporting
 *
 * Author
 * ======
 * Rob Stockley
 * New Zealand
 * robstockley@bigfoot.com
 *
 ******************************************************/

#include<stdio.h>
#include<sys/types.h>
#include<stdlib.h>
#include<string.h>
#include<dirent.h>  /* I read that some systems might not have this */

#define TRUE	1
#define FALSE	0

/*
 * Results tree is a linked list of linked lists of linked
 * lists....well you get the picture.
 */

typedef struct _bookmark {
  char 			*name;
  struct _bookmark 	*next;
  char 			*url;
} _bookmark;

typedef struct _folder {
  char			*name;
  struct _folder 	*next;
  struct _folder 	*child;
  struct _bookmark 	*bookmarks;
} _folder;

typedef _bookmark	*_book_p;
typedef _folder		*_fold_p;

/***************************************************
 * begins_in_url
 * =============
 * returns true if its char array argument starts
 * with "URL=", returns FALSE otherwise
 */
int begins_in_url(char *text) {
  char	mask[5] = "URL=";
  int	n;
  
  for (n = 0; n < 4; n++) {
    
    if (text[n] != mask[n]) {
      
      return FALSE;
    }
  }
  
  /* if we get to here then the mask matches */
  return TRUE;
}

/***************************************************
 * extract_url
 * ===========
 * returns a pointer to a char array that contains 
 * the internet address part of its string argument.
 */
char *extract_url(char *text) {
  char	*address = NULL;
  int	n;
  
  /* idiot check the argument */
  if (begins_in_url(text) == FALSE) {
    
    /* this shouldn't happen */
    fprintf(stderr, "Error: text argument passed to extract_url corrupt\n");
    exit(TRUE);
  }
  
  if ((address = (char *)malloc(strlen(text) - 3)) == NULL) {
    
    fprintf(stderr, "Error: malloc failed to allocate space for internet address\n");
    exit(TRUE);
  }
  
  /* copy the address but remove any linefeeds found */
  for (n = 0; n < (strlen(text) - 4); n++) {
    
    address[n] = text[n + 4];
  }

  /* add the terminating NULL.  We need to overwrite any nuisance
   * line feeds that DOS might have left 
   */
  if (address[n - 1] == '\x0d') {

    address[n - 1] = '\0';
  } else {

    address[n] = '\0';
  }
  
  return address;
}

/***************************************************
 * grab_line
 * =========
 * grabs the next line of input from its file pointer
 * argument and returns a pointer to it.  If the file
 * contains no more data then NULL is returned.
 *
 * Using char *fgets(char *s, int size, FILE *stream)
 * Need to be aware that the newline is returned with
 * the resultant string. We should return NULL if the 
 * EOF is found.
 */
char *grab_line(FILE *fp) {
  char	*bp;

  if ((bp = (char *)malloc(128)) == NULL) {

    fprintf(stderr, "Error: malloc failed to allocate space for stream buffer\n");
    exit(TRUE);
  }

  if ((bp = fgets(bp, 128, fp)) == NULL) {

    /* we've found the end of the file or encountered
     * an error.  Either way we're going to scrap the 
     * rest of this file stream so return NULL.
     */
    return NULL;
  }
  
  /* need to trim off the newline if it's there */
  if (bp[strlen(bp) - 1] == '\n') {

    /* snip */
    bp[strlen(bp) - 1] = '\0';
  }

  /* not sure if we could find an EOF but check anyway */
  if (bp[strlen(bp) - 1] == EOF) {

    /* snip */
    bp[strlen(bp) - 1] = '\0';
  }
  
  return bp;
}

/***************************************************
 * path_is_dir
 * ===========
 * checks to make sure path is a directory, it terminates
 * in a '/' character and that it can be opened and read
 */
int path_is_dir(char *path) {
  DIR	*dp = NULL;
  
  /* first check for a terminating '/' */
  if (path[(strlen(path) - 1)] != '/') {
    
    return FALSE;
  }
  
  /* next see if we can open the path */
  if ((dp = opendir(path)) != NULL) {
    
    closedir(dp); /* to be tidy */
    return TRUE;
  } else {
    
    return FALSE;
  }
}

/***************************************************
 * path_is_url
 * ===========
 * checks to make sure path matches the mask *.url, and
 * that it can be opened for input.
 */
int path_is_url(char *path) {
  char	mask[5] = ".url";
  int	offset, n;
  FILE	*fp = NULL;
  
  /* firstly do an idiot check on the length of path. 6 is the 
   * minimum length including the terminating null. This would 
   * mean the bookmark name was a single character
   */
  if (strlen(path) < 5) {
    
    return FALSE;
  }
  
  /* now determine the offset */
  offset = strlen(path) - 4;
  
  /* next do the checking */
  for (n = 0; n < 5; n++) {
    
    if (path[offset + n] != mask[n]) {
      
      return FALSE;
    }
  }
  
  /* if we got to here then the filename is cool but we
   * still need to see if we can open it.
   */
  if ((fp = fopen(path, "r")) == NULL) {
    
    /* couldn't open file so fail */
    return FALSE;
  } else {
    
    fclose(fp); /* to clean up */
  }
  
  /* if we get to here then the file has passed */
  return TRUE;
}

/* **************************************************
 * parse_for_url
 * ===========
 * Expects path to be a valid bookmark file (this is checked).
 * The file is parsed for a line beginning with 'URL='. It is
 * expected that the remainder of that line will be something 
 * like 'http:....' or 'ftp:....' and be a valid internet 
 * address (we won't check for stale addresses). The address in 
 * copied and a pointer to it is returned.  If the file can't be 
 * opened the program halts since only valid paths should get to 
 * this stage. If an EOF is reached before finding an appropriate 
 * address then a NULL is returned.  Leave the calling function to 
 * decide what to do here.
 */
char *parse_for_url(char *path) {
  FILE	*fp = NULL;
  char	*input_line = NULL;
  char	*new_url = NULL;

  /* firstly check our path */
  if (path_is_url(path) == FALSE) {
    
    /* this shouldn't happen so terminate */
    fprintf(stderr, "Error: bad path passed to extract_url\n");
    exit(TRUE);
  }
  
  /* initialise the file pointer */
  if ((fp = fopen(path, "r")) == NULL) {
    
    /* shouldn't happen so terminate */
    fprintf(stderr, "Error: parse_for_url couldn't open target file\n");
    exit(TRUE);
  }
  
  while ((input_line = grab_line(fp)) != NULL) {
    
    /* see if line begins in 'URL=' */
    if (begins_in_url(input_line) == TRUE) {
      
      /* don't need file anymore */
      fclose(fp);
      
      /* extract the internet address */
      if ((new_url = extract_url(input_line)) == NULL) {
	
				/* perhaps the bookmark file was corrupt */
	return NULL;
      } else {
	
	return new_url;
      }
    }
  }
  
  /* if we get to here then we didn't find the url */
  fclose(fp); /* to clean up */
  return NULL;
}

/***************************************************
 * get_folder_name
 * ===============
 * Expects a valid path as it's argument (this isn't 
 * checked).  Function takes this path and extracts the 
 * rightmost portion that should be the name of the deepest
 * folder on that path.  The name (if found) is copied into
 * a new space and a pointer to it is returned.  If the
 * name can't be found, or the path is too short to contain 
 * a name, then NULL is returned.  Leave it up to the calling
 * procedure to decide what to do.
 */
char *get_folder_name(char *path) {
  int	start, end, n;
  char	*name = NULL;
  
  /* first parse path to find start and end */
  for (end = strlen(path) - 1; ((path[end] != '/') && (end >= 1)); end--) ;
  if (end < 1) {
    
    /* something's wrong with path */
    return NULL;
  } else {
    
    /* make end point to the last character of the name */
    end--;
  }
  
  /* next parse for the start */
  for (start = (end - 1); ((path[start] != '/') && (start >= 0)); start--) ;
  
  /* make start point to the first character of the name */
  start++;
  
  /* check that we didn't have two '/' together */
  if (start > end) {
    
    return NULL;
  }
  
  /* create space for new name */
  if ((name = (char *)malloc(end - start + 2)) == NULL) {
    
    fprintf(stderr, "Error: malloc failed to create space for folder name\n");
    exit(TRUE);
  }
  for (n = 0; start + n <= end; n++) {
    
    name[n] = path[start + n];
  }
  
  /* add the terminating NULL character */
  name[n] = '\0';
  
  return name;
}

/***************************************************
 * get_bookmark_name
 * =================
 * Expects a valid path as it's argument (this isn't 
 * checked).  Function takes this path and extracts the 
 * rightmost portion that should be the name of the bookmark 
 * file. The name (if found) is copied into a new space and 
 * the '.url' bit is chopped off.  A pointer to it is returned.  
 * If the name can't be discerned then NULL is returned.  
 * Leave it up to the calling procedure to decide what to do.
 */
char *get_bookmark_name(char *path) {
  int	start, end, n;
  char	*name = NULL;
  
  /* first parse path to find start and end */
  for (end = strlen(path) - 1; ((path[end] != '.') && (end >= 1)); end--) ;
  if (end < 1) {
    
    /* something's wrong with path */
    return NULL;
  } else {
    
    /* make end point to the last character of the name */
    end--;
  }
  
  /* next parse for the start */
  for (start = (end - 1); ((path[start] != '/') && (start >= 0)); start--) ;
  
  /* make start point to the first character of the name */
  start++;
  
  /* check that we didn't have something silly like '/.url' */
  if (start > end) {
    
    return NULL;
  }
  
  /* create space for new name */
  if ((name = (char *)malloc(end - start + 2)) == NULL) {
    
    fprintf(stderr, "Error: malloc failed to create space for bookmark name\n");
    exit(TRUE);
  }
  for (n = 0; start + n <= end; n++) {
    
    name[n] = path[start + n];
  }
  
  /* add the terminating NULL character */
  name[n] = '\0';
  
  return name;
}

/***************************************************
 * merge_path
 * ==========
 * takes a valid path and joins target onto the end
 * of path adding a separating '/' if necessary. Function
 * returns a pointer to the new target. No checking is
 * done. When allocating space we'll include an extra
 * char in case we were called by merge_dir_path.  That
 * will save another call to malloc.
 */
char *merge_path(char *path, char *target) {
  char	*new_path = NULL;
  int	space;
  int	need_slash = FALSE;

  space = strlen(path) + 1;

  if (path[space - 2] != '/') {

    /* allocate extra char for '/' character */
    space++;
    need_slash = TRUE;
  }

  space += strlen(target) + 1;

  /* now we don't need to add the extra space for the possible
   * terminating '/' since we've counted two '\0' characters
   * so far.
   */

  if ((new_path = (char *)malloc(space)) == NULL) {

    fprintf(stderr, "Error: malloc failed to allocate space for new path\n");
    exit(TRUE);
  }

  if ((new_path = (char *)strcpy(new_path, path)) == NULL) {

    /* something's gone wrong */
    fprintf(stderr, "Error: something's gone wrong copying path\n");
    exit(TRUE);
  }

  if (need_slash == TRUE) {
    
    new_path[strlen(new_path) + 1] = '\0';
    new_path[strlen(new_path)] = '/';
  }

  /* idiot check the length of new_path */
  if ((strlen(new_path) - strlen(path)) > 1) {
   
    /* Oops! That didn't work */
    fprintf(stderr, "Error: new_path length is wrong\n");
    exit(TRUE);
  }

  if ((new_path = (char *)strcat(new_path, target)) == NULL) {

    /* something's gone wrong */
    fprintf(stderr, "Error: something's gone wrong copying target\n");
    exit(TRUE);
  }

  return new_path;
}

/***************************************************
 * merge_dir_path
 * ==============
 * calls merge_path initially then adds a terminating '/'
 * to the result as if the target were a directory. Function
 * returns a pointer to the new target.
 */
char *merge_dir_path(char *path, char *target) {
  char	*new_path;
  
  if ((new_path = merge_path(path, target)) == NULL) {
    
    fprintf(stderr, "Error: call to merge_path failed\n");
    exit(TRUE);
  }
  
  if ((new_path = strcat(new_path, "/")) == NULL) {
    
    fprintf(stderr, "merge_dir_path failed to add trailing '/'\n");
    exit(TRUE);
  }
  
  return new_path;
}

/***************************************************
 * prindent
 * ========
 * print out the required amount of leading whitespace
 */
void prindent(int indent) {
  int count;
  
  for (count = 0; count != indent; count++) fprintf(stdout, "  ");
}


/***************************************************
 * print_bookmarks
 * ===============
 * print the list of bookmarks given as its argument
 */
void print_bookmarks(_book_p bookmark, int indent) {
  _book_p	tm;

  for (tm = bookmark; tm != NULL; tm = tm->next) {

    prindent(indent);
    fprintf(stdout, "<DT><A HREF=\"%s\" ADD_DATE=\"931573536\" LAST_VISIT=\"0\" LAST_MODIFIED=\"0\">%s</A>\n", tm->url, tm->name);
  }
}

/***************************************************
 * print_folder
 * ============
 * recursively parse the folder passed as an argument
 * and create the HTML output.  Indent is passed to make 
 * the resulting HTML source more readable to humans
 * like me. Netscape does this by default so why shouldn't 
 * we.  Nothing to return.
 */
void print_folder(_fold_p folder, int indent) {
  _fold_p	tf;

  prindent(indent);
  fprintf(stdout, "<DL><p>\n");
  indent++;
  prindent(indent);
  fprintf(stdout, "<DT><H3 FOLDED ADD_DATE=\"931571800\">%s</H3>\n", folder->name);

  if (folder->child != NULL) {

    for (tf = folder->child; tf->next != NULL; tf = tf->next) {

      print_folder(tf, indent + 1);
    }
  }

  if (folder->bookmarks != NULL) {

    prindent(indent);
    fprintf(stdout, "<DL><p>\n");
    print_bookmarks(folder->bookmarks, indent + 1);
    prindent(indent);
    fprintf(stdout, "</DL><p>\n");
  }

  indent--;
  prindent(indent);
  fprintf(stdout, "</DL><p>\n");
}

/***************************************************
 * mk_bookmark
 * ===========
 * Expects a pointer to a valid path as its argument.
 * Searches the file named in path for a valid URL entry
 * then creates a new bookmark node and returns a pointer
 * to it.  If we can't find a valid URL in the file then
 * the file might be either corrupt, incomplete, or not
 * really a bookmark.  In that case return a NULL so the
 * calling function can ignore it.
 */
_book_p mk_bookmark(char *path){
  _book_p	new_bookmark = NULL;
  char		*bookmark_url = NULL;
  char		*bookmark_name = NULL;
  
  /* firstly we're expecting a valid path but lets check anyway */
  if (path_is_url(path) == FALSE) {
    
    /* we were expecting good data here so something is wrong */
    fprintf(stderr, "Error: path given to mk_bookmark is corrupt\n");
    exit(TRUE);
  }
  
  /* if we get to here then our path must be good */
  
  /* now try to get the url entry from the file given in path */
  if ((bookmark_url = parse_for_url(path)) == NULL) {
    
    /* obviously couldn't find the URL entry, perhaps
     * it wasn't there to be found. Anyway return a NULL */
    return NULL;
  }
  
  /* extract the bookmark name from the path */
  if ((bookmark_name = get_bookmark_name(path)) == NULL) {
    
    /* this shouldn't happen */
    fprintf(stderr, "Error: failed trying to get bookmark name\n");
    exit(TRUE);
  }
  
  /* create the new bookmark */
  if ((new_bookmark = (_book_p)malloc(sizeof(_bookmark))) == NULL) {
    
    /* malloc has failed */
    fprintf(stderr, "Error: malloc failed to allocate space for new bookmark\n");
    exit(TRUE);
  }
  
  /* fill in the bookmark, initialize NULL's just in case */
  new_bookmark->name = bookmark_name;
  new_bookmark->next = NULL;
  new_bookmark->url = bookmark_url;
  
  /* if we get to here then everything must be ok */
  return new_bookmark;
}

/***************************************************
 * mk_folder
 * =========
 * Expects a pointer to a valid path as its argument.
 * Allocates space for one new folder then adds the 
 * contents of path to that folder.  It is called 
 * recursively to add sub-folders.  Function returns 
 * a pointer to the new folder.
 */
_fold_p mk_folder(char *path) {
  _fold_p	new_folder = NULL;
  _fold_p	child_keeper = NULL;
  _book_p	bookmark_keeper = NULL;
  char		*folder_name = NULL;
  DIR		*dp = NULL;
  struct dirent	*dent = NULL;
  char		*target_path = NULL;
  
  /* first check that path is valid */
  if (path_is_dir(path) == FALSE) {
    
    /* this shouldn't happen so terminate */
    fprintf(stderr, "Error: path given to mk_folder is not a valid dir\n");
    fprintf(stderr, "Error: path = '%s'\n", path);
    exit(TRUE);
  }
  
  /* extract the folder name from the path */
  if ((folder_name = get_folder_name(path)) == NULL) {
    
    /* if this fails then something's not working right */
    fprintf(stderr, "Error: couldn't glean name from path\n");
    exit(TRUE);
  }
  
  /* create the new folder */	
  if ((new_folder = (_fold_p)malloc(sizeof(_folder))) == NULL) {
    
    fprintf(stderr, "Error: malloc failed to create space for new folder\n");
    exit(TRUE);
  } 
  
  /* set up new folder.  Initialise NULL's to be sure */
  new_folder->name = folder_name;
  new_folder->next = NULL;
  new_folder->child = NULL;
  new_folder->bookmarks = NULL;
  
  /* Next scan directory contents. We could scan for 
   * directories and bookmarks in the same loop, but 
   * that looks untidy so do them one at a time.
   */
  
  /* scan for directories */
  if ((dp = opendir(path)) == NULL) {
    
    /* this shouldn't happen */
    fprintf(stderr, "Error: failed to open path for scanning folders\n");
    exit(TRUE);
  }
  while ((dent = readdir(dp)) != NULL) {
    
    /* firstly ignore . and .. entries */
    if (dent->d_name[0] != '.') {
      
      /* need to create path to new entry */
      if ((target_path = merge_dir_path(path, dent->d_name)) == NULL) {
	
				/* time to check code if this happens */
	fprintf(stderr, "Error: failed to merge path with dent contents\n");
	exit(TRUE);
      }
      
      /* if this is a valid directory then add it to the tree */
      if (path_is_dir(target_path) == TRUE) {
	
				/* store the address of the children */
	child_keeper = new_folder->child;
	
				/* add new folder node */
	if ((new_folder->child = mk_folder(target_path)) == NULL) {
	  
	  /* something's gone wrong */
	  fprintf(stderr, "Error: mk_folder returned NULL ?????");
	  exit(TRUE);
	}
	new_folder->child->next = child_keeper;
      }
    }
  }
  closedir(dp);  /* just to be tidy */
  
  /* scan for bookmarks */
  if ((dp = opendir(path)) == NULL) {
    
    /* this definitely shouldn't happen */
    fprintf(stderr, "Error: failed to open path for scanning bookmarks\n");
    exit(TRUE);
  }
  while ((dent = readdir(dp)) != NULL) {
    
    /* firstly ignore . and .. entries */
    if (dent->d_name[0] != '.') {
      
      /* need to create path to new entry */
      if ((target_path = merge_path(path, dent->d_name)) == NULL) {
	
				/* time to check code if this happens */
	fprintf(stderr, "Error: failed to merge path with dent contents\n");
	exit(TRUE);
      }
      
      /* if this is a valid bookmark then add it to the tree */
      if (path_is_url(target_path) == TRUE) {
	
				/* store the address of the children */
	bookmark_keeper = new_folder->bookmarks;
	
				/* add new folder node */
	if ((new_folder->bookmarks = mk_bookmark(target_path)) == NULL) {
	  
	  /* perhpas couldn't extract url info from file so ignore */
	  new_folder->bookmarks = bookmark_keeper;
	} else {
	  
	  /* update the new bookmark */
	  new_folder->bookmarks->next = bookmark_keeper;
	}
      }
    }
  }
  closedir(dp);
  
  /* if we get to here then everything should have gone ok */
  return new_folder;
}

int main(int argc, char *argv[]) {
  _fold_p	results_tree = NULL;
  char		*target_path = NULL;
  
  /* firstly check we have the correct number of args */
  if (argc != 2) {
    
    fprintf(stderr, "Usage: %s ie_favorites_dir\n", argv[0]);
    fprintf(stderr, "   eg: '%s /dos/windows/favorites/'\n", argv[0]);
    exit(TRUE);
  }
  
  /* next fix up the target path */
  target_path = merge_path(argv[1], "");
  if (path_is_dir(target_path) == FALSE) {
    
    /* try adding a '/' */
    target_path = merge_dir_path(argv[1],"");
    if (path_is_dir(target_path) == FALSE) {
      
      /* can't resolve path so print something useful */
      fprintf(stderr, "Error: couldn't resolve the path you supplied\n");
      fprintf(stderr, "Error: check that the path exists and that you have\n");
      fprintf(stderr, "Error: permission to access it.  Perhaps you forgot\n");
      fprintf(stderr, "Error: to mount your Windows drive.\n");
      exit(TRUE);
    }
  } /* if we get to here then the path is good */	
  
  /* parse the target path and build the results tree */
  if ((results_tree = mk_folder(target_path)) == NULL) {
    
    /* definitely something wrong here */
    fprintf(stderr, "Error: failed to create results tree at top level\n");
    exit(TRUE);
  }
  
  /* parse the results tree and output HTML source, indent starts at 0 */
  print_folder(results_tree, 0);
  
  /* don't forget to return something! */
  return FALSE;
}

/* end of fav2bmk.c */

